import logging
from laboro.error import LaboroError
from laboro.module.validator import Validator
from laboro.module.specification import Specification

MODULES = {}


class Module:
  """The ``laboro.module.Module`` object is the base object for all Laboro external modules. It defines the basic methods to manipulate the module specification and arguments.

  Arguments:
    specification: Optional, a ``laboro.module.specification.Specification`` object representing the expected arguments. If no ``Specification`` is given, then the module will not accept any argument.

    args: Optional, a dictionary representing the given arguments for the module instance.

  Raises:
    ``laboro.error.LaboroError``: When ``args`` is not a dictionary.

  The following code snippet illustrates how to create a Laboro external module derived from the ``laboro.module.Module`` base class.

  ..  code-block:: python

    from laboro.modules.module import Module
    from laboro.error import LaboroError

    class ExampleModule(Module):
      def __init__(self, args=None):
        spec = {"arg1": {"required": True,
                         "type": int,
                         "implied": ["arg3"],
                         "excluded": []},
                "arg2": {"required": False,
                         "type": str,
                         "implied": [],
                         "excluded": []},
                "arg3": {"required": False,
                         "type": bool,
                         "implied": [],
                         "excluded": ["arg2"]}
        super().__init__(specification=spec, args=args)
        <some init code>

      def example_method(self):
        try:
          <some code>
        except Exception as err:
          raise LaboroError(f"[SomeError] Some error occurred: [{err.__class__.__name__}]: {err}") from err
  """

  @property
  def args(self):
    return self._args

  @args.setter
  def args(self, args):
    if isinstance(args, dict):
      self._args = args
    else:
      raise LaboroError(f"[ModuleArgError]: Expected a dict as args, received {type(args)}")

  def __init__(self, specification=None, args=None):
    self.specification = Specification(specification)
    if args is None:
      args = dict()
    self.args = args
    self.validate_args()

  def validate_args(self):
    validator = Validator(specification=self.specification)
    try:
      return validator.validate(self.args)
    except LaboroError as err:
      logging.critical(str(err))
      raise err

  def get_arg_value(self, arg):
    """Get the value of the specified argument.

    Arguments:
      arg: The argument to get the value from.

    Returns:
      The value of the searched argument. The type of the return value depend of the value of the searched argument.

    Raises:
      ``laboro.error.LaboroError``: When the specified argument is not known to the module.
    """
    if arg in self.args:
      return self.args[arg]
    raise LaboroError(f"UnknownModuleArg: Module argument not found: {arg}")

  def get_arg_value_as_string(self, arg):
    """Get the value of the specified argument as a string.

    Arguments:
      arg: The argument to get the value from.

    Returns:
      ``str``: The string representation of value of the searched argument.

    Raises:
      ``laboro.error.LaboroError``: When the specified argument is not known to the module.
    """
    if arg in self.args:
      return str(self.args[arg])
    raise LaboroError(f"UnknownModuleArg: Module argument not found: {arg}")
