import logging
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from laboro.error import LaboroError
from laboro.module.specification.schema import SPECIFICATION_SCHEMA


class Specification:
  """The ``laboro.module.validator.specification`` convert a dict object representing a laboro.module.Module arguments specification into a consistent object.

  Arguments:
    spec: Optional, a dict object representing the module arguments specification. If the ``spec`` argument is not given or its value is equal to ``None``, the ``laboro.module.Module`` object using this ``Specification`` will not accept any argument.

  Raises:
    ``laboro.error.LaboroError``: When the given dict does not respect the expected format.

  When given, the ``spec`` dict is expected to conform to the the following the ``laboro.module.specification.schema.SPECIFICATION_SCHEMA`` *jsonschema*.

  ..  code-block:: python

      SPECIFICATION_SCHEMA = {
          "$schema": "https://json-schema.org/draft/2020-12/schema",
          "$id": "laboro.module.specification.schema",
          "title": "Laboro module specification schema",
          "description": "Laboro module specification schema",
          "type": "object",
          "properties": {
              "args": {
                  "type": "array",
                  "uniqueItems": True,
                  "items": {
                      "type": "object",
                      "properties": {
                          "name": {
                              "type": "string",
                          },
                          "required": {
                              "type": "boolean",
                          },
                          "type": {
                              "type": "string",
                          },
                          "excluded": {
                              "type": "array",
                              "uniqueItems": True,
                              "items": {
                                  "type": "string",
                              },
                          },
                          "implied": {
                              "type": "array",
                              "uniqueItems": True,
                              "items": {
                                  "type": "string",
                              },
                          },
                          "secret": {
                              "type": "boolean",
                          },
                      },
                      "required": ["name",
                                  "required",
                                  "type",
                                  "excluded",
                                  "implied",
                                  "secret"],
                      "additionalProperties": False,
                  }
              },
          },
          "required": ["args"],
          "additionalProperties": False,
      }


  **Note**:  The argument ``type`` property is a string representing any built-in *Python* type or object class.

  For a more precise type validation, the expected object class should be imported prior the ``Specification`` instantiation and **must be available in the the module scope**.

  Example:

  The following specification indicates:

  - ``first_arg`` is mandatory, expects a boolean value, can not be used alongside ``second_arg``, implies no other argument and its value is not secret.
  - ``second_arg`` is not mandatory, expects a string, can not be used alongside ``first_arg``, implies that ``third_arg`` must be set and its value is secret.
  - ``third_arg`` is not mandatory, expects an unspecified object, can be used alongside all other arguments, implies none of them and its value is not secret.

  ..  code-block:: python

      from laboro.module.validator import Specification

      my_spec = {"args": [
                  {"name": "arg1",
                   "required": True,
                   "type": "bool",
                   "excluded": ["second_arg"],
                   "implied": [],
                   "secret": False},
                  {"name": "second_arg",
                   "required": False,
                   "type": "str",
                   "excluded": ["first_arg"],
                   "implied": ["third_arg"],
                   "secret": True},
                  {"name": "arg3",
                   "required": False,
                   "type": "MyClass",
                   "excluded": [],
                   "implied": [],
                   "secret": False},
                  ]
                }
      spec = Specification(my_spec)
  """

  @property
  def spec(self):
    """Get the spec as a dict."""
    return self._spec

  @spec.setter
  def spec(self, spec):
    """Set the spec"""
    if spec is None or self._validate(spec):
      self._spec = spec
    else:
      raise LaboroError("[SpecFormatError] Specification is not valid.")

  def __init__(self, spec=None):
    self.spec = spec

  def _validate(self, spec):
    try:
      validate(instance=spec, schema=SPECIFICATION_SCHEMA)
      return True
    except ValidationError as err:
      logging.critical(f"[{err.__class__.__name__}] {err}")
      return False
