import builtins
from laboro.error import LaboroError


class Validator:
  """The purpose ``laboro.module.validator.Validator`` class is to validate arguments passed to module against the argument specification provided by the module

  Arguments:
    specification: A ``laboro.module.specification.Specification`` object representing the module arguments specification.
  """

  def __init__(self, specification=None):
    self.specification = specification

  def _validate_required(self, arguments):
    required = sorted([arg["name"] for arg in list(filter(lambda arg: arg["required"], self.specification.spec["args"]))])
    return sorted([arg for arg in arguments if arg in required]) == required

  def _validate_type(self, arg, value):
    arg_type = [aarg["type"] for aarg in self.specification.spec["args"] if aarg["name"] == arg][0]
    arg_type = getattr(builtins, arg_type)
    if not isinstance(value, arg_type):
      msg = {"type": "BadArgTypeError",
             "message": f"Expected type {arg_type} / received {type(value)} for {arg}."}
      raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def _validate_implied(self, arg, arguments):
    try:
      implied = sorted([aarg["implied"] for aarg in list(filter(lambda arg: arg["implied"], self.specification.spec["args"])) if aarg["name"] == arg])[0]
    except IndexError:
      implied = list()
    if sorted([iarg for iarg in arguments if iarg in implied]) != implied:
      msg = {"type": "MissingImpliedArgTypeError",
             "message": f"All args of {implied} are required by {arg}."}
      raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def _validate_excluded(self, arg, arguments):
    try:
      excluded = sorted([aarg["excluded"] for aarg in list(filter(lambda arg: arg["excluded"], self.specification.spec["args"])) if aarg["name"] == arg])[0]
    except IndexError:
      excluded = list()
    if len([earg for earg in arguments if earg in excluded]) > 0:
      msg = {"type": "ExcludedArgTypeError",
             "message": f"None of {' or '.join(excluded)} can be used alongside {arg}."}
      raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def validate(self, arguments):
    """Validate the given arguments against the ``self.spec`` specification.

    Arguments:
      arguments: A dict object representing the arguments and their values that will be passed to the module.

    Returns:
      ``bool``: True, if arguments validates against the specification.

    Raises:
      ``laboro.error.LaboroError``: When the given arguments does not validate against the specification or when the specification is not defined.
    """
    msg = {"type": "UnexpectedError", "message": "An unexpected error occurred while validating module arguments."}
    if self.specification.spec is not None:
      if self._validate_required(arguments):
        for arg, value in arguments.items():
          if arg in [arg["name"] for arg in self.specification.spec["args"]]:
            self._validate_type(arg, value)
            self._validate_implied(arg, arguments)
            self._validate_excluded(arg, arguments)
          else:
            msg = {"type": "UnknownArgError",
                   "message": f"Unknown argument: {arg}."}
            raise LaboroError(f"[{msg['type']}] {msg['message']}")
        return True
      raise LaboroError("[MissingRequiredArgError] Missing required argument.")
    raise LaboroError("[UndefinedSpecError] No argument specification defined.")
