import os
import re
import time
import uuid
import logging
from laboro.error import LaboroError
from laboro.config.manager import Manager as ConfigMgr
from laboro.module.manager import Manager as ModuleMgr
from laboro.error.handler import Handler as ErrorHandler
from laboro.module import MODULES
from laboro.history import History
from laboro.workspace import Workspace


class Workflow:
  """The ``laboro.workflow.Workflow`` object is the main class for the workflow representation.
  It load all configuration needed, set alk objects such History, Vault and workspace and run according to its configuration.

  The Workflow object provides a runtime context that will handle log, history, vault, and workspace, etc.

  Arguments:
    name: A string representing the workflow name.
    logger: A laboro.logger.manager.Manager instance.

  Returns:
    ``laboro.workflow.Workflow``: A Workflow object.

  ..  code-block:: python

      from laboro.workflow import Workflow
      from laboro.logger.manager import Manager as LogMgr

      logmgr = LogMgr()
      logmgr.set_default()
      with Workflow(name="my_workflow", logger=logmgr) as wkf:
        wkf.run()
        ...
  """
  def __init__(self, name, logger):
    self.name = name
    self.logger = logger
    self.session = str(uuid.uuid4())
    self.configmgr = ConfigMgr("/etc/laboro/laboro.yml")
    workspacedir = self.configmgr.get_parameter("main",
                                                "$.laboro.workspacedir")
    histdir = self.configmgr.get_parameter("main",
                                           "$.laboro.histdir")
    workflowdir = self.configmgr.get_parameter("main",
                                               "$.laboro.workflowdir")
    logdir = self.configmgr.get_parameter("main",
                                          "$.laboro.logdir")
    self.logger.add_file_handler(logdir, self.name, self.session)
    self.configmgr.workflow_config = os.path.join(workflowdir,
                                                  f"{self.name}.yml")
    self.history = History(filename=os.path.join(histdir, f"{self.name}.db"),
                           workflow=self.name,
                           session=self.session,
                           params=self.configmgr.workflow_config)
    self.workspace = Workspace(workspacedir=workspacedir,
                               workflow=self.name,
                               session=self.session)
    self.modulemgr = ModuleMgr()
    self.vault = logging.getLogger().vault
    self.vault.clear()

  def __enter__(self):
    self.history.__enter__()
    self.install_packages()
    return self

  def __exit__(self, kind, value, traceback):
    self.workspace.delete()
    self.history.__exit__(kind, value, traceback)
    exit_code = 0
    if kind == SystemExit:
      if value.code is not None:
        exit_code = value.code
    elif kind is not None:
      exit_code = kind.__name__
    logging.info(f"[+] Workflow ended with code: {exit_code}")
    ErrorHandler().handle_error(kind, value, traceback)
    self.logger.remove_file_handler(self.session)

  def install_packages(self):
    """Install all packages listed in the workflow configuration file.
    """
    packages = self.configmgr.get_parameter("workflow", "$.laboro.workflow.packages")
    reg = re.compile("^laboro-.*$")
    for pkg in packages:
      if re.match(reg, pkg) is not None:
        self.modulemgr.install_package(pkg)
      else:
        logging.critical(f"Package is not a valid Laboro package: {pkg}")
        raise LaboroError(f"InvalidPackageNameError: Module is not a valid Laboro package: {pkg}")

  def run(self):
    """Run the workflow."""
    logging.info(f"  [+] Running workflow: [{self.name}] {self.session}")
    time.sleep(1)
